"""SNAPstraction mesh/network ops"""

import binascii
import logging

import random

from tornado.concurrent import Future
from tornado.gen import coroutine, sleep, Return

from .node_cache import NodeCache

log = logging.getLogger(__name__)


@coroutine
def unicast_ping(gus, timeout, output_func, progress_cb=None, finished_cb=None):
    """Ping the mesh w/ Unicast vmstats"""

    data = {}

    with NodeCache() as cache:

        class nonlocal:
            pass

        nonlocal.found_count = 0
        nonlocal.last_addr = ''

        def on_vmStat(src, *args):
            """tellVmStat callback"""
            (word, d) = args

            word = word & 0xFFFF

            code = word & 0xFF
            hibyte = word >> 8

            # Make sure it's our tellVmStat...
            if code == 5:
                lq = hibyte
                data[src] = lq
                cache.add(src)
                nonlocal.found_count += 1
                nonlocal.last_addr = src

        gus.on_rpc('tellVmStat', on_vmStat)
        gus.mcast_rpc('vmStat', 5, timeout, group=1, ttl=timeout if timeout < 255 else 50)

        secs = 0
        while secs < timeout:
            yield sleep(1)
            if progress_cb:
                progress_cb(secs, nonlocal.found_count, nonlocal.last_addr)
            secs += 1

        if finished_cb:
            finished_cb()

    raise Return(data)


@coroutine
def dmcast_ping(gus, timeout, output_func, response_name=None, progress_cb=None, finished_cb=None):
    """Ping the mesh with directed multicast"""
    data = {}
    with NodeCache() as cache:
        if not response_name:
            response_name = "dmPing%06d" % random.randint(0, 999999)

        log.debug("Looking for response: %s", response_name)

        class nonlocal:
            pass

        nonlocal.found_count = 0
        nonlocal.last_addr = ''

        def on_response(src, *args):
            if src not in data:
                data[src] = args
                cache.add(src)
                nonlocal.found_count += 1
                nonlocal.last_addr = src

        gus.on_rpc(response_name, on_response)

        secs = 0
        while secs < timeout:
            yield gus.mcast_rpc('dmcastRpc', binascii.unhexlify(gus.address), 1, timeout, 50, response_name, 'ping', group=1, ttl=timeout if timeout < 255 else 20)
            yield sleep(1)
            if progress_cb:
                progress_cb(secs, nonlocal.found_count, nonlocal.last_addr)
            secs += 1

        if finished_cb:
            finished_cb()

    raise Return(data)


@coroutine
def listen(gus, output_func, _timeout=None):
    """Listen for packets that come our way"""
    def any_rpc(src, func, *args):
        output_func({'src': src, 'func': func, 'args': args})

    def any_data(src, data):
        output_func({'src': src, 'data': data})

    gus.on_any_rpc(any_rpc)
    gus.on_data(any_data)

    log.debug("Registered, waiting")

    if _timeout:
        yield sleep(_timeout)
    else:
        yield Future()


@coroutine
def listen_for_intercepted_output(gus, output_func, _timeout=None):
    """Listen for intercepted output from nodes"""
    def on_data(src, data):
        output_func({'node': src, 'packet': data})

    gus.on_data(on_data)

    if _timeout:
        yield sleep(_timeout)
    else:
        yield Future()
